unit uCore;

interface

uses
  Math, SysUtils;

const
  _maxlen = 3000;

type
  TLargeArr = array[1.._maxlen] of byte;

  TLargeNum = class(TObject) //   - 
  private
    Len: integer; // 
    Value: TLargeArr; //  -
  public
    constructor Create(n: integer); overload;
      //    
    constructor Create(n: TLargeNum); overload;
      //    
    procedure Clear();
      // 
    procedure AssignNumber(n: integer); overload;
    procedure AssignNumber(n: TLargeNum); overload;
    procedure AssignNumber(n: string); overload;
      //  
    function  Add(n: TLargeNum):TLargeNum; overload;
    function  Add(n: integer):TLargeNum; overload;
      // 
    function  Sub(n: TLargeNum):TLargeNum; overload;
    function  Sub(n: integer):TLargeNum; overload;
      // 
    function  MulByDigit(n: byte):TLargeNum;
      //   
    function  Mul(n: TLargeNum):TLargeNum; overload;
    function  Mul(n: integer):TLargeNum; overload;
      //   
    function  Fact2(): TLargeNum;
    function  Fact(): TLargeNum;
      //   (!)
    function  ShiftLeft(cnt: integer): TLargeNum;
      //  
    function  ShiftRight(cnt: integer): TLargeNum;
      //  
    function  TruncZeroes(): TLargeNum;
      //  
    function  Equ(Num2: TLargeNum): boolean; overload;
    function  Equ(Num2: integer): boolean; overload;
      //   
    function  GetDigit(apos: integer): byte;
      //  
    function  ToString(): WideString;
      //   
    function  ToInteger(): integer;
      //   
    function  GetLen(): integer;
      //  
  end;

  function CharToByte(c: char): byte;
  function ByteToChar(b: byte): char;

implementation

function IsDigit(c: char): boolean;
begin
  Result := (c >= '0') and (c <= '9');
end;

function CharToByte(c: char): byte;
begin
  if (not IsDigit(c)) then raise Exception.Create('Wrong digit');
  Result := byte(c)-48;
end;

function ByteToChar(b: byte): char;
begin
  Result := char(b + 48);
end;

{ TLargeNum }

function TLargeNum.Add(n: integer):TLargeNum;
var
  tmp: TLargeNum;
begin
  tmp := TLargeNum.Create(n);
  Result := Self.Add(tmp);
  tmp.Free();
end;

function TLargeNum.Add(n: TLargeNum):TLargeNum;
var
  Tail: byte;
  i: integer;
  a1, a2: byte;
  asum: byte;
  MaxLen: integer;
  TempArr: TLargeArr;
begin

  Tail := 0;
  MaxLen := max(Self.Len, n.Len);

  FillChar(TempArr , _maxlen, 0);

  for i := 1 to MaxLen do begin
    a1 := Self .GetDigit(i);
    a2 := n    .GetDigit(i);
    asum := (a1 + a2 + Tail) mod 10;
    Tail := (a1 + a2 + Tail) div 10;
    TempArr[i] := asum;
  end;

  TempArr[MaxLen + 1] := Tail;

  if (Tail > 0) then inc(MaxLen);

  Self.Value := TempArr;
  Self.Len := MaxLen;

  Result := Self;
end;


{procedure TLargeNum.AfterConstruction();
begin
  inherited;
//  Clear();
end;}

procedure TLargeNum.AssignNumber(n: integer);
var
  st: string;
  i: integer;
begin
  Clear();
  Str(n, st);
  Len := Length(st);
  for i := 1 to Len do Value[i] := CharToByte(st[Len + 1 -i]);
end;

procedure TLargeNum.AssignNumber(n: TLargeNum);
var
  i : integer;
begin
  Clear();
  Self.Len := n.Len;
  for i := 1 to Len do begin
    Value[i] := n.Value[i]; // todo: Move method
  end;
end;

procedure TLargeNum.AssignNumber(n: string);
var
  i: integer;
  aLen: integer;
begin
  aLen := Length(n);
  if (aLen < 1) then raise Exception.Create('Wrong digit');
  Len := aLen;
  for i := 1 to aLen do begin
    Value[i] := CharToByte(n[Len + 1 - i]);
  end;
end;

procedure TLargeNum.Clear;
begin
  Len := 1;
  FillChar(Value, _maxlen, 0);
end;

constructor TLargeNum.Create(n: TLargeNum);
begin
  inherited Create();
  AssignNumber(n);
end;

constructor TLargeNum.Create(n: integer);
begin
  inherited Create();
  AssignNumber(n);
end;

function TLargeNum.Equ(Num2: TLargeNum): boolean;
var
  i: integer;
begin

  Result := false;
  if (Self.Len <> Num2.Len) then exit;
  for i := 1 to Len do
    if Self.Value[i] <> Num2.Value[i] then exit;
  Result := true;
end;

function TLargeNum.Equ(Num2: integer): boolean;
var
  tmp: TLargeNum;
begin
  tmp := TLargeNum.Create(Num2);
  Result := Self.Equ(tmp);
  tmp.Free;
end;

function TLargeNum.Fact: TLargeNum;
var
  tmp: TLargeNum;
  i: integer;
begin

  tmp := TLargeNum.Create(1);

  if (Self.Equ(0)) then begin
    Result := tmp;
    exit;
  end;

  for i := 1 to Self.ToInteger do tmp.Mul(i);
  Result := tmp;

end;

function TLargeNum.Fact2: TLargeNum;
var
  tmp: TLargeNum;
  i: TLargeNum;
begin

  tmp := TLargeNum.Create(1);
  i := TLargeNum.Create(0);
  if (Self.Equ(i)) then begin
    Result := tmp;
    exit;
  end;

  repeat
    i.Add(1);
    tmp.Mul(i);
  until (Self.Equ(i));

  Result := tmp;
end;

function TLargeNum.GetDigit(apos: integer): byte;
begin
  Result := (Value[apos]);
end;

function TLargeNum.GetLen: integer;
begin
  Result := Len;
end;

function TLargeNum.Mul(n: integer):TLargeNum;
var
  tmp: TLargeNum;
begin
  tmp := TLargeNum.Create(n);
  Result := Self.Mul(tmp);
  tmp.Free;
end;

function TLargeNum.Mul(n: TLargeNum):TLargeNum;
var
  i: integer;
  Total: TLargeNum;
  dig: byte;
  ACopy: TLargeNum;

begin
  Total := TLargeNum.Create(0);
  ACopy := TLargeNum.Create(Self);

  for i := 1 to n.Len do begin
    dig := n.GetDigit(i);
    ACopy.AssignNumber(Self);
    ACopy.MulByDigit(dig);
    ACopy.ShiftLeft(i-1);
    Total.Add(ACopy);
  end;

  Self.AssignNumber(Total);
  Total.Free();
  ACopy.Free();
  Result := Self;
end;

function TLargeNum.MulByDigit(n: byte): TLargeNum;
var
  Tail: byte;
  i: integer;
  a: byte;
  amul: byte;
  TempArr: TLargeArr;
begin

  if (n=0) then begin
    Self.AssignNumber(0);
    Result := Self;
    exit;
  end;

  Tail := 0;
  FillChar(TempArr , _maxlen, 0);
  for i := 1 to Len do begin
    a := Self .GetDigit(i);
    amul := (a * n + Tail) mod 10;
    Tail := (a * n + Tail) div 10;
    TempArr[i] := amul;
  end;

  TempArr[Len + 1] := Tail;

  if (Tail > 0) then inc(Len);

  Self.Value := TempArr;
  Result := Self;
end;

function TLargeNum.ShiftLeft(cnt: integer): TLargeNum;
begin
  if (((Len = 1) and (Value[1] = 0)) or (cnt<1)) then begin

    Result := Self;
    exit;
  end;
  Move(Value[1], Value[cnt+1], Len);
  FillChar(Value[1], cnt, 0);
  Value[1] := 0;
  inc(Len, cnt);
  Result := Self;
end;

function TLargeNum.ShiftRight(cnt: integer): TLargeNum;
begin
  if (cnt >= Len) then begin
    AssignNumber(0);
    Result := Self;
    exit;
  end;

  Len := Len-cnt;
  Move(Value[cnt+1], Value[1], Len);
  Result := Self;
end;

function TLargeNum.Sub(n: integer):TLargeNum;
begin
  Result := nil; //  . ! :)
end;

function TLargeNum.Sub(n: TLargeNum):TLargeNum;
begin
  inherited;
  Result := nil; //  . ! :)
end ;

function TLargeNum.ToInteger: integer;
begin
  Result := StrToInt(Self.ToString);
end;

function TLargeNum.ToString(): WideString;
var
  st: string;
  i: integer;
begin
  SetLength(st, Len);
  for i := 1 to len do st[i] := ByteToChar(Value[Len + 1 - i]);
  Result := st;
end;

function TLargeNum.TruncZeroes: TLargeNum;
var
  cnt : integer;
begin
  cnt := 0;
  while (GetDigit(cnt+1) = 0) do inc(cnt);
  ShiftRight(cnt);
  Result := Self;
end;

end.
